package main

import (
	"bytes"
	"errors"
	"fmt"
	"net"
	"os"
	"os/exec"
	"path/filepath"
	"strings"
	"syscall"

	"io/ioutil"

	"gopkg.in/alecthomas/kingpin.v2"
	"gopkg.in/yaml.v2"

	"strconv"

	errs "github.com/pkg/errors"
	"github.com/vishvananda/netlink"
)

type cmdList []string

func (i *cmdList) Set(value string) error {
	*i = append(*i, value)
	return nil
}

func (i *cmdList) String() string {
	return ""
}

func (i *cmdList) IsCumulative() bool {
	return true
}

func CmdList(s kingpin.Settings) (target *[]string) {
	target = new([]string)
	s.SetValue((*cmdList)(target))
	return
}

var (
	app               = kingpin.New("octopus-outside", "nspawn Bootstrapper.").Author("Sangbum Kim<sangbumkim@amuz.es>")
	verbosePtr        = app.Flag("verbose", "Enable verbose mode.").Short('v').Bool()
	clearPtr          = app.Flag("clear", "clear os environment.").Short('p').Bool()
	ifBindTypePtr     = app.Flag("if-bind-type", "choose virtual ethernet bind type(ipvlan,macvlan).").Default("ipvlan").Short('b').Enum("ipvlan", "macvlan")
	isFindSubnetPtr   = app.Flag("subnet-scan", "find subnet mode. if not specify, use default network scan mode.").Short('t').Default("false").Bool()
	macAddrPtr        = app.Flag("mac-address", "Specify MAC address of macvlan mode. if not specify, use random.").Short('a').String()
	nspawnPathPtr     = app.Flag("nspawn-path", "systemd-nspawn location.").Short('s').String()
	containerHomePtr  = app.Flag("container-home", "container residence path").Default("/container").Short('d').String()
	configFileNamePtr = app.Flag("config-file", "config file name").Default("settings.yml").HintOptions("FILENAME").Short('c').String()
	nodeNamePtr       = app.Arg("nodeName", "nodename to spawn.").Required().String()
	bindIfNamePtr     = app.Arg("bindInterface", "interface name to bind.").Required().String()
	cmdListArgsPtr    = CmdList(app.Arg("cmd", "execute command rather than boot process"))
)

type mountPoint struct {
	Name     string `yaml:"name"`
	Map      string `yaml:"map"`
	Readonly bool   `yaml:"readonly"`
}

func (mp mountPoint) Option(nodePath string) (option string, err error) {
	var mountPointPath string
	if filepath.IsAbs(mp.Name) {
		mountPointPath = mp.Name
	} else {
		mountPointPath = filepath.Clean(filepath.Join(nodePath, "volume", mp.Name))
	}

	if _, err = os.Stat(mountPointPath); os.IsNotExist(err) {
		err = errors.New(fmt.Sprintf("mount point  %s is not exist", mountPointPath))
		return
	}

	var buffer bytes.Buffer
	if mp.Readonly {
		buffer.WriteString("--bind-ro=")
	} else {
		buffer.WriteString("--bind=")
	}
	buffer.WriteString(mountPointPath)
	buffer.WriteRune(':')
	buffer.WriteString(mp.Map)
	option = buffer.String()
	return
}
func (mp mountPoint) String() string {
	var buffer bytes.Buffer
	buffer.WriteString(mp.Name)
	if mp.Readonly {
		buffer.WriteString("(readonly)")
	}
	buffer.WriteString("->")
	buffer.WriteString(mp.Map)
	return buffer.String()
}

type config struct {
	MoundPoint []mountPoint `yaml:"mount-point"`
}

func (c config) String() string {
	var buffer bytes.Buffer
	buffer.WriteString("config:\n")
	buffer.WriteString("  mount-point:\n")
	for _, entry := range c.MoundPoint {
		buffer.WriteString("    - ")
		buffer.WriteString(entry.String())
		buffer.WriteRune('\n')
	}
	return buffer.String()
}

// LookupIP looks up host using the local resolver.
// It returns an array of that host's IPv4 and IPv6 addresses.
func getIP(host string) (ip net.IP, err error) {
	var addrs []net.IP
	addrs, err = net.LookupIP(host)
	if err != nil {
		return
	}
	return addrs[0], nil
}

func getCidr(network net.IPNet) (cidr uint64, err error) {
	netSlice := strings.Split(network.String(), "/")
	if len(netSlice) < 2 {
		err = errors.New("cannot get netmask")
		return
	}
	cidr, err = strconv.ParseUint(netSlice[len(netSlice)-1], 10, 64)
	return
}
func getSubnetIPInfo(ip net.IP, verbose bool) (gw net.IP, cidr uint64, linkname string, err error) {
	// netlink.RouteList
	var links []netlink.Link
	links, err = netlink.LinkList()
	if err != nil {
		return
	}

	if verbose {
		fmt.Println("scan system network interfaces:")
	}

	for _, link := range links {
		attrs := link.Attrs()

		if verbose {
			fmt.Printf(" interface %s : ", attrs.Name)
		}
		if (attrs.Flags & net.FlagUp) == 0 {
			if verbose {
				fmt.Printf(" not up -- abandon!\n")
			}
			continue
		} else {
			if verbose {
				fmt.Printf(" up!\n")
			}
		}
		if verbose {
			fmt.Println("   getting routing tables ")
		}
		var routes []netlink.Route
		routes, err = netlink.RouteList(link, netlink.FAMILY_V4)
		if err != nil {
			continue
		}
		var (
			matchedRoute *netlink.Route
		)
		for _, route := range routes {

			if verbose {
				fmt.Printf("     routing table %s\n", route.String())
			}

			if route.Dst == nil {
				if verbose {
					fmt.Printf("      - default gateway : no one!\n")
				}
			} else if route.Dst.Contains(ip) {
				if verbose {
					fmt.Printf("      - matched routing role : %s\n", route.String())
				}
				matchedRoute = &route
				break
			} else if verbose {
				fmt.Printf("      - no subnet gateway\n")
			}
		}

		if matchedRoute == nil {
			if verbose {
				fmt.Println("    - cannot getting routing rule")
				fmt.Println("  abandon this interface ")
			}
			continue
		}
		if cidrData, err := getCidr(*matchedRoute.Dst); err == nil {
			cidr = cidrData
			if verbose {
				fmt.Printf("     interface have a subnet routing rule. gateway: %s", matchedRoute)
				fmt.Println(" ok! matched ")
			}
			gw, cidr, linkname = matchedRoute.Src, cidrData, attrs.Name
			break
		} else if verbose {
			fmt.Printf("    - cannot getting netmask : %s", err)
			fmt.Println("  abandon this interface ")
		}
	}
	return
}
func getOutboundIPInfo(verbose bool) (gw net.IP, cidr uint64, linkname string, err error) {
	// netlink.RouteList
	var links []netlink.Link
	links, err = netlink.LinkList()
	if err != nil {
		return
	}

	if verbose {
		fmt.Println("scan system network interfaces:")
	}
	for _, link := range links {
		attrs := link.Attrs()

		if verbose {
			fmt.Printf(" interface %s : ", attrs.Name)
		}
		if (attrs.Flags & net.FlagUp) == 0 {
			if verbose {
				fmt.Printf(" not up -- abandon!\n")
			}
			continue
		} else {
			if verbose {
				fmt.Printf(" up!\n")
			}
		}

		if verbose {
			fmt.Println("   getting routing tables ")
		}
		var routes []netlink.Route
		routes, err = netlink.RouteList(link, netlink.FAMILY_V4)
		if err != nil {
			continue
		}
		var (
			matchedGw   net.IP
			matchedCidr uint64
			cidrOk      = false
		)
		for _, route := range routes {

			if verbose {
				fmt.Printf("     routing table %s\n", route.String())
			}
			if route.Dst == nil {
				if verbose {
					fmt.Printf("      - default gateway\n")
				}
				matchedGw = route.Gw
				break
			} else if verbose {
				fmt.Printf("      - no default gateway\n")
			}
		}
		if matchedGw == nil {
			if verbose {
				fmt.Println("     interface have not a default routing rule.")
			}
			continue
		} else if verbose {
			fmt.Printf("     interface have a default routing rule. gateway: %s", matchedGw.String())
		}

		if verbose {
			fmt.Println("   getting networks ")
		}
		var networks []netlink.Addr
		networks, err = netlink.AddrList(link, netlink.FAMILY_V4)
		if err != nil {
			fmt.Println("    - cannot getting networks ")
			continue
		}
		for _, network := range networks {
			if verbose {
				fmt.Printf("     network %s\n", network.String())
			}
			if network.Contains(matchedGw) {
				if verbose {
					fmt.Printf("      - default network\n")
				}
				if cidrData, err := getCidr(*network.IPNet); err == nil {
					matchedCidr = cidrData
					cidrOk = true
					break
				} else if verbose {
					fmt.Printf("    - cannot getting netmask : %s", err)
				}
			} else if verbose {
				fmt.Printf("      - no default network\n")
			}
		}
		if cidrOk {
			gw, cidr, linkname = matchedGw, matchedCidr, attrs.Name
			if verbose {
				fmt.Println(" ok! matched ")
			}
			break
		} else if verbose {
			fmt.Println("  abandon this interface ")
		}
	}
	if gw == nil {
		err = errors.New("default gateway or cidr not found!")
	}

	return
}

func checkInterfaceValid(bindIfName string) (err error) {
	var link netlink.Link
	link, err = netlink.LinkByName(bindIfName)

	if err != nil {
		return
	}
	attrs := link.Attrs()
	if (attrs.Flags & net.FlagUp) == 0 {
		err = errors.New("bind interface had been deactivated")
	}
	return
}

func checkNodeBasicDirectory(containerHomePath string, nodeName string, configFileName string) (nodePath string, nodeConfigPath string, err error) {
	var containerHomeInfo os.FileInfo
	if containerHomeInfo, err = os.Stat(containerHomePath); os.IsNotExist(err) {
		err = errors.New(fmt.Sprintf("container home path %s is not exist", containerHomePath))
		return
	} else if !containerHomeInfo.IsDir() {
		err = errors.New(fmt.Sprintf("container home path %s is not a directory", containerHomePath))
		return
	}

	nodePath = filepath.Join(containerHomePath, nodeName)

	var nodePathInfo os.FileInfo
	if nodePathInfo, err = os.Stat(nodePath); os.IsNotExist(err) {
		err = errors.New(fmt.Sprintf("node path %s is not exist", nodePath))
		return
	} else if !nodePathInfo.IsDir() {
		err = errors.New(fmt.Sprintf("node path %s is not a directory", nodePath))
		return
	} else {
		subDirNames := [...]string{"merge", "work", "delta"}
		for _, subDirName := range subDirNames {
			subPath := filepath.Join(nodePath, subDirName)

			var subPathInfo os.FileInfo
			if subPathInfo, err = os.Stat(subPath); os.IsNotExist(err) {
				err = errors.New(fmt.Sprintf("sub path %s is not exist", subPath))
				return
			} else if !subPathInfo.IsDir() {
				err = errors.New(fmt.Sprintf("sub path %s is not a directory", subPath))
				return
			}
		}
	}
	nodeConfigPath = filepath.Join(containerHomePath, nodeName, configFileName)

	var nodeConfigPathInfo os.FileInfo
	if nodeConfigPathInfo, err = os.Stat(nodeConfigPath); os.IsNotExist(err) || nodeConfigPathInfo.IsDir() {
		nodeConfigPath = ""
	}

	return

}

func LoadConfig(path string) (config config, err error) {
	var (
		source []byte
	)
	if path == "" {
	} else if source, err = ioutil.ReadFile(path); err != nil {
	} else if err = yaml.Unmarshal(source, &config); err != nil {
	}
	return
}

func main() {
	_, err := app.Parse(os.Args[1:])
	var (
		verbose        = *verbosePtr
		ifBindType     = *ifBindTypePtr
		macAddr        = strings.TrimSpace(*macAddrPtr)
		clear          = *clearPtr
		nodeName       = *nodeNamePtr
		isFindSubnet   = *isFindSubnetPtr
		bindIfName     = *bindIfNamePtr
		configFileName = *configFileNamePtr
		cmdListArgs    = *cmdListArgsPtr
		nspawnPath     string
		progress       = -1
	)

	defer func() {
		if r := recover(); r != nil {
			fmt.Fprintln(os.Stderr, r)
			os.Exit(progress)
		}
	}()
	if err != nil {
		panic(err)
	}

	progress = 1
	containerHomePath, err := filepath.Abs(*containerHomePtr)
	if err != nil {
		panic(err)
	}

	progress++
	nodePath, nodeConfigPath, err := checkNodeBasicDirectory(containerHomePath, nodeName, configFileName)
	if err != nil {
		panic(err)
	} else if verbose {
		var nodeConfigPathDesc = "not used"
		if nodeConfigPath != "" {
			nodeConfigPathDesc = nodeConfigPath
		}
		fmt.Printf(`sane container environment:
  - container home : %s
  - node home : %s
  - node config : %s
`, containerHomePath, nodePath, nodeConfigPathDesc)
	}
	config, err := LoadConfig(nodeConfigPath)
	if err != nil {
		panic(err)
	} else if verbose {
		fmt.Printf(config.String())
	}

	progress++

	if nspawnPathPtr != nil && *nspawnPathPtr != "" {
		nspawnPath = *nspawnPathPtr
	} else if nspawnPathData, err := exec.LookPath("systemd-nspawn"); err != nil {
		panic(errs.Wrap(err, "cannot find systemd-nspawn"))
	} else {
		nspawnPath = nspawnPathData
		if verbose {
			fmt.Printf("systemd-nspawn found : %s\n", nspawnPath)
		}
	}

	progress++
	err = checkInterfaceValid(bindIfName)
	if err != nil {
		panic(errs.Wrap(err, "error in checkInterfaceValid"))
	}
	progress++
	nodeAddr, err := getIP(nodeName)

	if err != nil {
		panic(errs.Wrap(err, "error in getIP"))
	}

	progress++
	var (
		gw       net.IP
		cidr     uint64
		linkname string
	)
	if isFindSubnet {
		gw, cidr, linkname, err = getSubnetIPInfo(nodeAddr, verbose)
	} else {
		gw, cidr, linkname, err = getOutboundIPInfo(verbose)

	}
	if err != nil {
		panic(errs.Wrap(err, "error in network info"))
	}

	if verbose {
		fmt.Println("network configuration:")
		fmt.Printf("  - bind interface{%s} : valid\n", bindIfName)
		fmt.Printf("  - default routing info : ip %s/%d link %s\n", gw.String(), cidr, linkname)
		fmt.Printf("  - node{%s} IP address : %s\n", nodeName, nodeAddr.String())
		fmt.Printf("  - bind interface type : %s\n", ifBindType)
	}
	option := []string{
		"--quiet",
		"--link-journal=try-host",
		"--notify-ready=true"}
	switch ifBindType {
	case "macvlan":
		option = append(option, fmt.Sprintf("--network-macvlan=%s", bindIfName))
		if len(macAddr) > 0 {
			option = append(option, fmt.Sprintf("--setenv=MAC_ADDR=%s", macAddr))
		}
	case "ipvlan":
		option = append(option, fmt.Sprintf("--network-ipvlan=%s", bindIfName))
	default:
		panic(errors.New("unknown bind type"))
	}

	option = append(option, fmt.Sprintf("--setenv=IF_TYPE=%s", ifBindType))
	option = append(option, fmt.Sprintf("--setenv=IP_ADDR=%s/%d", nodeAddr.String(), cidr))
	option = append(option, fmt.Sprintf("--setenv=GATEWAY_ADDR=%s", gw.String()))
	option = append(option, fmt.Sprintf("--machine=%s", nodeName))
	option = append(option, fmt.Sprintf("--directory=%s", filepath.Join(nodePath, "merge")))
//        option = append(option, "--keep-unit", "--register=yes", "--settings=override", "--private-users=1354956800:65536", "--private-users-chown")
        option = append(option, "--keep-unit", "--register=yes", "--settings=override")

	for _, mountPoint := range config.MoundPoint {
		if mountOption, err := mountPoint.Option(nodePath); err != nil {
			panic(errs.Wrap(err, "error in mountOption"))
		} else {
			option = append(option, mountOption)
		}
	}
	if len(cmdListArgs) == 0 {
		if verbose {
			fmt.Println("startup type: using systemd")
		}
		option = append(option, "--boot")
	} else {
		if verbose {
			fmt.Println("startup type: using command")
		}
		for _, cmdArg := range cmdListArgs {
			option = append(option, fmt.Sprintf(" %s", cmdArg))
		}
	}
	var env []string
	if clear {
		env = []string{}
	} else {
		env = os.Environ()
		if verbose {
			fmt.Println("os environment:")
			for _, segment := range env {
				fmt.Println("  ", segment)
			}
		}
	}
	if verbose {
		fmt.Println("systemd-nspawn spawning:")
		fmt.Println(nspawnPath, " \\")
		for _, segment := range option {
			fmt.Println("  ", segment, " \\")
		}
		fmt.Println("  ;")
		fmt.Println("bye!")
	}
	if err := syscall.Exec(nspawnPath, option, env); err != nil {
		panic(errs.Wrap(err, "error in systemd-nspawn execution"))
	}
}
